from typing import Tuple, Iterable
from agent import Agent
from random import shuffle


class Rule:
    '''
    General interface with only one method which says whether the first player wins or not.
    '''

    def whoWins(self, hands: Tuple[Iterable[str], Iterable[str]]) -> int:
        '''
        Returns 0 if the first hand winds, 1 otherwise.

        :param hands: Tuple[Iterable[str], Iterable[str]] 
        :return: int 
        '''
        pass


class Sum(Rule):
    def __init__(self):
        self.strToValue = {"ACE": 14,
                           "KING": 13,
                           "OBER": 12,
                           "UNTER": 11,
                           "TEN": 10,
                           "NINE": 9,
                           "EIGHT": 8,
                           "SEVEN": 7,
                           "SIX": 6,
                           "FIVE": 5,
                           "FOUR": 4,
                           "THREE": 3,
                           "TWO": 2
                           }

    def value(self, card: str) -> int:
        return self.strToValue[card.split("_")[1]]

    def sum(self, hand: Iterable[str]) -> int:
        return sum(map(self.value, hand))

    def whoWins(self, hands: Tuple[Iterable[str], Iterable[str]]) -> bool:
        return 0 if sum(self.value(card) for card in hands[0]) >= sum(self.value(card) for card in hands[1]) else 1


class Oracle:
    def __init__(self, handSize: int = 2, cards: int = 10, colors: Tuple[str] = ['BLUE', 'RED'], shapes: Tuple[str] = ['BELLS', 'HEARTS', 'ACORNS', 'LEAVES'], rule: Rule = Sum(), samples: int = 10):
        self.handSize: int = handSize
        self.cards: int = cards
        self.colors: int = colors
        self.shapes: int = shapes
        self.rule: Rule = rule
        self.samples: int = samples

        valueToNumber = {14: "ACE", 13: "KING", 12: "OBER", 11: "UNTER", 10: "TEN", 9: "NINE", 8: "EIGHT", 7: "SEVEN", 6: "SIX", 5: "FIVE", 4: "FOUR", 3: "THREE", 2: "TWO"}

        if cards + 2 not in valueToNumber:
            raise ValueError("at most 13 cards are available")

        cards = ["{}_{}_OF_{}".format(color, valueToNumber[value + 2], shape) for color in colors for value in range(0, cards) for shape in shapes]

        if 2 * handSize > len(cards):
            raise ValueError("there are not enough cards for this hand size")

        self.hands = []

        for _ in range(0, samples):
            shuffle(cards)
            self.hands.append([cards[0:handSize], cards[handSize:2 * handSize]])

    def __iter__(self):
        for idx in range(0, len(self.hands)):
            self.idx = idx
            yield self.hands[idx]

    def submitAnswer(self, answer: int) -> int:
        '''
        Given the answer (saying whether the first (0) or the second (1) player has won), the oracle returns a reward for this answer (prediction) by returning 0 and -1 for correct and incorrect prediction respectively.
          
        :param answer: int 
        :return: int
        '''
        return 0 if answer == self.rule.whoWins(self.hands[self.idx]) else -1


def main():
    samples = 10
    oracle = Oracle(handSize=2, cards=4, shapes=['BELLS', 'HEARTS'], colors=['RED'], samples=samples, rule=Sum())
    agent = Agent()

    for sample in oracle:
        answer = agent.receiveSample(sample)
        isCorrectAnswer = oracle.submitAnswer(answer)
        agent.receiveReward(isCorrectAnswer)
        print("agent's hypothesis\t{}".format(agent.getHypothesis()))

    print("agent's hypothesis\t{}".format(agent.getHypothesis()))


if __name__ == "__main__":
    main()
