import pymzn
import itertools
from logic import *
from collections import defaultdict
from typing import Tuple, Dict, Set, Callable
import time


def addSpecialSymbol(literal: Literal) -> Literal:
    prefix = "~~" if literal.positive else "~~!"
    predicate = Predicate(prefix + literal.atom.predicate.name, literal.atom.predicate.arity)
    return Literal(Atom(predicate, literal.atom.terms), positive=True)


def termsToDict(expression, mapping: Dict[Term, int] = None) -> Dict[Term, int]:
    if None is mapping:
        mapping = {}
    if isinstance(expression, Constant) or isinstance(expression, Variable):
        if expression not in mapping:
            mapping[expression] = len(mapping) + 1
    else:
        if isinstance(expression, CompoundTerm) and (expression not in mapping):
            mapping[expression] = len(mapping) + 1
        for e in expression:
            mapping = termsToDict(e, mapping)
    return mapping


def encodeToCSP(query: Clause, world: Clause) -> Tuple[str, Dict[int, Term], Dict[str, Variable]]:
    # only for FOL
    # returns either a string encoding of the subsumption task, or None if query cannot subsume the world
    worldTermsToNumber: Dict[Constant, int] = termsToDict(world)
    literalsPerPredicate: Dict[Predicate, Set[Literal]] = defaultdict(set)
    for l in world:
        literalsPerPredicate[l.getPredicate()].add(l)
        if not l.positive:
            raise ValueError("only non-negated literals are allowed for the world variable")

    # variable introduction
    cspVarToFOLVar: Dict[str, Variable] = {}
    folVarToCSPVar: Dict[Variable, str] = {}  # this mapping is here because we need to overcome restriction of variable names in the CSP solver (e.g. semicolons are definitely not allowed to be part of the name)
    problem = ""
    for v in query.getVariables():
        cspVariable = "v{}".format(len(cspVarToFOLVar))
        cspVarToFOLVar[cspVariable] = v
        folVarToCSPVar[v] = cspVariable
        # here, we need to add type restriction if it is introduced in the framework
        problem += "var 1..{}: {};\n".format(len(worldTermsToNumber), cspVariable)  # if string of a variable contains its type (similar as in TreeLiker), the variables has to be renamed here

    constraints = ""
    for l in query:
        if l.isGround():
            if ((l.positive and l not in world) or ((not l.positive) and l.negation() in world)):
                return None, None, None  # if a query's l is ground but is not in the world, then the query cannot subsume the world (similarly for negative literal)
            continue

        if (not l.positive) and len(literalsPerPredicate[l.getPredicate()]) < 1:
            continue

        mappable = []
        literalToLiteralConnection = "==" if l.positive else "!="
        for potential in literalsPerPredicate[l.getPredicate()]:

            possibleUnification = unifyFirstWithSecond(l, potential)
            if None is possibleUnification:
                continue

            innerConstraintConnection = " /\\ " if l.positive else " \/ "
            mappable.append("( " + innerConstraintConnection.join("{} {} {}".format(folVarToCSPVar[src], literalToLiteralConnection, worldTermsToNumber[img]) for src, img in possibleUnification.items()) + " )")

        if len(mappable) < 1 and not l.positive:  # there are no literals on which this on can be mapped
            continue

        if len(mappable) < 1:  # there is no literal in world to which l can be mapped
            return None, None, None

        if l.positive:
            constraints += "constraint\t{};\n".format(" \/ ".join(mappable))
        else:
            constraints += "\n".join(["constraint\t{};".format(forbidden) for forbidden in mappable]) + "\n"

    # append constraints
    problem += "\n" + constraints
    problem += "\n\nsolve satisfy;"

    intToTerm = {value: key for key, value in worldTermsToNumber.items()}
    return problem, intToTerm, cspVarToFOLVar


def findSubstitutions(query: Clause, world: Clause, allSolutions=True) -> Iterable[Dict[Variable, Term]]:
    '''
    By default it search for all substitutions (all_solutions=True).
    
    So far, no types are considered. TBD in future.
    
    :param query: Clause
    :param world: Clause
    :param allSolutions: bool 
    :return: Iterable[Dict[Variable,Term]] 
    '''

    # special (ground) case
    if len(query.getVariables()) < 1:
        for l in query:
            if (l.positive and l not in world) or ((not l.positive) and l.negation() in world):
                return []
        return [{}]  # although it looks weird, it actually means that there is one empty substitutions, i.e. a very trivial one

    # general case
    start = time.time()
    cspFormulation, intToTerm, cspVarToFOLVar = encodeToCSP(query, world)
    encodingTime = time.time() - start

    # print(cspFormulation, intToTerm)

    if None is cspFormulation:
        return []

    start = time.time()
    solutions = pymzn.minizinc(cspFormulation, all_solutions=allSolutions)
    cspTime = time.time() - start

    # print("encoding\t{}\ncsp     \t{}".format(encodingTime, cspTime))
    # print(cspFormulation)

    # transfer solution to FOL representation
    substitutions = [{cspVarToFOLVar[src]: intToTerm[img] for src, img in substitution.items()} for substitution in solutions]
    return substitutions


def subsumes(query: Clause, world: Clause) -> bool:
    '''
    Given a world and a query, the method decides whterh the query theta-subsume the world.
    :param query: Clause 
    :param world: Clause
    :return: bool
    '''
    query = Clause(map(addSpecialSymbol, query))
    world = Clause(map(addSpecialSymbol, world))

    # print("query & world\n{}\n{}".format(query, world))   # debug
    substitutions = findSubstitutions(query, world, allSolutions=False)
    return len(substitutions) > 0


def getHead(clause: Clause, polarity=True) -> Literal:
    '''
    Returns the only literal with given polarity. If no such literal is presented, it raises an exception.
    
    :param clause: Clause
    :param polarity: bool
    :return: Literal 
    '''
    positiveLiterals = list(l for l in clause if polarity == l.positive)
    if len(positiveLiterals) != 1:
        raise ValueError()
    return positiveLiterals[0]


def leastHerbrandModel(rules: Iterable[Clause], world: Iterable[Atom]) -> Iterable[Atom]:
    '''
    Given an iterable of clauses, i.e. definite rules, and a world, i.e. list of true facts, it returns the least Herbrand model, i.e. all positive literals that derived from the input parameters 
     
    :param rules: Iterable[Clause]
    :param world: Iterable[Atom]
    :return: Clause
    '''
    existentiallyQuantifiedDefiniteRules: Tuple[Clause, Literal] = set()
    for clause in rules:
        if isDefiniteClause(clause):
            existentiallyQuantifiedDefiniteRules.add((clause.flipSings(), getHead(clause, polarity=True)))
        else:
            raise ValueError("this method takes only definite rules, but given\t{}".format(clause))
    facts: Clause = Clause([Literal(a) for a in world])
    change = True
    while change:
        change = False
        for c, head in existentiallyQuantifiedDefiniteRules:
            substitutions = findSubstitutions(c, facts)
            change |= len(substitutions) > 0
            facts = facts.extend([head.substitute(s) for s in substitutions])

    return [literal.atom for literal in facts]


def reduce(clause: Clause) -> Clause:
    '''
    Returns a reduced clause for the given one.
    
    :param clause: Clause
    :return: Clause 
    '''
    reduced = clause
    for literal in clause:
        if literal not in reduced:
            continue  # speeding-up hack
        shorther = Clause(l for l in reduced if l != literal)
        if subsumes(reduced, shorther):
            reduced = shorther
    return reduced


def lggByTaxonomic(t1: Constant, t2: Constant, antiunifier: Callable[[Term, Term], Term]) -> Term:
    # if the two input constants are the same, then returns the constant
    # if they have common lowest common ancestor, then it is returned
    # else it returns a variable
    if t1 == t2:
        return t1
    # place for taxonomical information insertion
    return antiunifier(t1, t2)


def lggTerms(t1: Term, t2: Term, antiunifier: Callable[[Term, Term], Term]) -> Term:
    if isinstance(t1, Constant) or isinstance(t2, Constant):
        return lggByTaxonomic(t1, t2, antiunifier)
    if isinstance(t1, CompoundTerm) and isinstance(t2, CompoundTerm) and t1.functor == t2.functor:
        return CompoundTerm(t1.functor,
                            (lggTerms(it1, it2, antiunifier) for it1, it2 in zip(t1.terms, t2.terms)))
    return antiunifier(t1, t2)


def lgg(alpha: Clause, beta: Clause, reduceResult: bool = True) -> Clause:
    '''
    Returns (reduced) lgg of clauses alpha and beta and reduces.
    
    :param alpha: Clause
    :param beta: Clause
    :param reduceResult: bool 
    :return: Clause 
    '''
    cache = {}

    def antiunifier(t1: Term, t2: Term) -> Variable:
        if isinstance(t1, Constant) and t1 == t2:
            return t1
        key = (t1, t2)
        if key not in cache:
            cache[key] = Variable("v" + str(len(cache)))  # well, this should be parametrized by constVar or varConst parameter :))
        return cache[key]

    general = Clause(Literal(Atom(l1.getPredicate(), (lggTerms(t1, t2, antiunifier) for t1, t2 in zip(l1.atom.terms, l2.atom.terms))), positive=l1.positive)
                     for l1, l2 in itertools.product(alpha, beta) if
                     l1.getPredicate() == l2.getPredicate() and l1.positive == l2.positive)

    if reduceResult:
        general = reduce(general)
    return general


def tValue(clause: Clause) -> int:
    '''
    Return the maximal t value, i.e. the number of occurrences of predicates and terms' symbols, of literals in the given clause. 
    
    The method handles function free logic only.
    
    :param clause: Clause 
    :return: int 
    '''
    # function-free
    for c in clause:
        for l in c:
            if isinstance(l, CompoundTerm):
                raise NotImplementedError("method tValue is implemented for function-free logic but input is of form:\t{}".format(clause))

    return max(l.atom.arity + 1 for l in clause)


def generateST(s: int, t: int, predicates: Set[Predicate], rangeRestricted: bool = True) -> List[Clause]:
    '''
    Returns list of all possible (range-restricted) st-function and constant-free clauses.  
    
    :param s: int 
    :param t: int
    :param predicates: Set[Predicate] 
    :param rangeRestricted:  bool
    :return: List[Clause]
    '''
    # it is a function-free generator of rr-st clauses
    retVal = list()

    beam: Set[Clause] = set()
    beam.add(Clause([]))
    for length in range(0, s):
        children = list()

        for c in beam:
            variables = c.getVariables()
            for predicate in predicates:
                newVariables = freshVariables(variables, predicate.arity)
                newAtom = Atom(predicate, newVariables)
                for polarity in [True, False]:
                    newLiteral = Literal(newAtom, positive=polarity)
                    for permutation in itertools.permutations(unionSets([variables, newVariables]), r=predicate.arity):
                        substituted = newLiteral.substitute({newVariables[idx]: permutation[idx] for idx in range(0, predicate.arity)})
                        if substituted in c:
                            continue
                        refinement = c.append(substituted)

                        # print("{}".format(refinement))

                        if (not isRangeRestricted(refinement)) or len(reduce(refinement)) != len(c) + 1 or tValue(refinement) > t:
                            # print("\t{}\t{}\t{}".format(not isRangeRestricted(refinement), len(reduce(refinement)) != len(c) + 1, tValue(refinement)))
                            continue
                        children.append(refinement)
        unique = []
        for i in range(0, len(children)):
            c1 = children[i]
            original = True
            for j in range(i + 1, len(children)):
                c2 = children[j]
                if subsumes(c1, c2) and subsumes(c2, c2):
                    original = False
                    break
            if original:
                unique.append(c1)

        beam = children
        retVal.extend(unique)

    return retVal
